#include "stdarg.h"
#include "slog_impl.h"

int SLOG_OpenOutputStream( SLOG_STREAM *slog, const char *filename )
{
    SLOG_InitGlobalData( slog );
    if ( SLOG_Open( slog, name, "wb" ) != SLOG_SUCCESS )
        return SLOG_FAIL;
    if ( SLOG_WriteDirHdr( slog ) != SLOG_SUCCESS )
        return SLOG_FAIL;
    if ( SLOG_WritePreview( slog ) != SLOG_SUCCESS )
        return SLOG_FAIL;
    if ( SLOG_WriteProfile( slog, SLOG_PROFILENAME ) != SLOG_SUCCESS )
        return SLOG_FAIL;
    if ( SLOG_ReadWriteRecDefs( slog, SLOG_RECDEFFILE ) != SLOG_SUCCESS )
        return SLOG_FAIL;
    if ( SLOG_WriteInitFrameDir( slog ) != SLOG_SUCCESS )
        return SLOG_FAIL;
    SLOG_STM_Init( slog )

    return SLOG_SUCCESS;
}

int SLOG_AppendRecord( SLOG_STREAM             *slog,
                       const SLOG_irectype_t   irectype,
                       const SLOG_etype_t      etype,
                       const SLOG_starttime_t  starttime,
                       const SLOG_duration_t   duration,
                       const SLOG_bebit_t      bebits[2],
                       const SLOG_nodeID_t     node_id,
                       const SLOG_cpuID_t      cpu_id,
                       const SLOG_threadID_t   thread_id,
                       const SLOG_where_t      where,
                       const SLOG_assoc_t      assoc,
                       ... )
{
    /*  Local Variables  */
    va_list ap;

    SLOG_intvlrec_t intvlrec;
    int expected_Nargs;
    int sizeof_N_MPIargs;
    int sizeof_IntvlRec_in_file;
    int total_irec_in_frame;
    int Nbytes_written;

    int ii;
    int ierr;
    
#if defined( DEBUG )
    long slogfile_loc_cur;
#endif

    /*  Assemble all the input arguments as a SLOG_intvlrec_t, intvlrec  */
    intvlrec.irectype  = irectype ;
    intvlrec.etype     = etype    ;
    intvlrec.starttime = starttime;
    intvlrec.duration  = duration ;
    intvlrec.bebits[1] = bebits[1];
    intvlrec.bebits[2] = bebits[2];
    intvlrec.thread_id = thread_id;
    intvlrec.node_id   = node_id  ;
    intvlrec.cpu_id    = cpu_id   ;
    intvlrec.where     = where    ;
    intvlrec.assoc     = assoc    ;

    intvlrec.N_MPIargs = SLOG_Buf_NumOfArgs( slog, irectype );

  /*
      Read in the MPI call's corresponding argument list 
      as determined by input irectype
  */
    if ( intvlrec.N_MPIargs > 0 ) {
      /*  Compute the memory needed in interval record and Allocate it  */
        sizeof_N_MPIargs   = intvlrec.N_MPIargs * SLOG_typesz[ Narg_t ];
        intvlrec.args      = ( SLOG_MPIarg_t * ) malloc( sizeof_N_MPIargs );

        va_start( ap, assoc );
        for ( ii = 0; ii < intvlrec.N_MPIargs; ii++ ) {
          /*
             Needs Exceptional Handling to check if the number of the
             actual arguments is less than than expected, intvlrec.N_MPIargs.
          */
            intvlrec.args[ ii ] = va_arg( ap,  SLOG_MPIarg_t );
        };
        va_end( ap );
    }
    else {
        sizeof_N_MPIargs   = 0;     /*  Init it, in case irec.N_MPIargs = 0  */
        intvlrec.args      = NULL;
    }

  /*  Compute the size of the current Interval Record  */
    sizeof_IntvlRec_in_file = SLOG_typesz[ min_IntvlRec ]
                            + sizeof_N_MPIargs
                            * SLOG_typesz[ Narg_t ];

  /*  Check if writing the input interval record to the buffer is possible  */
    new_buf_Nbytes_in_file = SLOG_typesz[ FrameHdr ]
                           + ( slog->inc_bbuf ).Nbytes_in_file
                           + ( slog->cur_bbuf ).Nbytes_in_file
                           + ( slog->out_bbuf ).Nbytes_in_file
                           + sizeof_IntvlRec_in_file;

      /*  i.e.  check if new buffer size is too big for the buffer  */
    if ( new_buf_Nbytes_in_file > SLOG_BUFFER_UNIT ) {

      /*
          Locate interval records in Incoming and Current Bbufs
          that fulfull the criteria ( exceed the Frame's endtime )
          are copied to the Temporary Bbuf
      */
        ierr = SLOG_Buf_UpdateBBuf_endtime( &( slog->tmp_bbuf ),
                                            &( slog->inc_bbuf ),
                                            slog->frame_dir_rec );
        if ( ierr != SLOG_SUCCESS ) {
            fprintf( errfile, __FILE__":SLOG_AppendRecord() - "
                              "SLOG_Buf_UpdateBBuf_endtime(Tmp_Bbuf,Inc_Bbuf) "
                              "fails!!\n" );
            fflush( errfile );
            return SLOG_FAIL;
        };
        ierr = SLOG_Buf_UpdateBBuf_endtime( &( slog->tmp_bbuf ), 
                                            &( slog->cur_bbuf ),
                                            slog->frame_dir_rec );
        if ( ierr != SLOG_SUCCESS ) {
            fprintf( errfile, __FILE__":SLOG_AppendRecord() - "
                              "SLOG_Buf_UpdateBBuf_endtime(Tmp_Bbuf,Inc_Bbuf) "
                              "fails!!\n" );
            fflush( errfile );
            return SLOG_FAIL;
        };


      /*  Update the buffer's Frame Directory Record  */
        ierr = SLOG_STM_UpdateFrameDirRec( slog );
        if ( ierr != SLOG_SUCCESS ) {
            fprintf( errfile, __FILE__":SLOG_AppendRecord() - "
                              "SLOG_STM_UpdateFrameDirRec( slog ) "
                              "fails!!\n" );
            fflush( errfile );
            return SLOG_FAIL;
        };

      /*  Update the buffer's Frame Header  */ 
        total_irec_in_frame = ( slog->inc_bbuf ).count_irec
                            + ( slog->cur_bbuf ).count_irec
                            + ( slog->out_bbuf ).count_irec;
        SLOG_STM_UpdateFrameHdr( slog, total_irec_in_frame ,
                                 ( slog->frame_dir_rec ).ftr2framehdr
                                 + ( slog->inc_bbuf ).Nbytes_in_file ,
                                 ( slog->frame_dir_rec ).ftr2framehdr
                                 + ( slog->inc_bbuf ).Nbytes_in_file
                                 + ( slog->cur_bbuf ).Nbytes_in_file ,
                                 SLOG_fptr_NULL );

        Nbytes_written = 0;

#if defined( DEBUG )
        slogfile_loc_cur = ftell( slog->fd );
        if ( slogfile_loc_cur != ( slog->frame_dir_rec ).ftr2framehdr ) {
            fprintf( errfile, __FILE__":SLOG_AppendRecord() - "
                              "Inconsistency in File Pointers before calling "
                              "SLOG_STM_WriteFrameHdr() at line "__LINE__"\n" );
            fprintf( errfile, "       ""- slogfile_loc_cur = %d, but "
                              "( slog->frame_dir_rec ).ftr2framehdr = %d\n",
                              slogfile_loc_cur,
                              ( slog->frame_dir_rec ).ftr2framehdr );
            fflush( errfile );
        };
#endif

      /*  Output buffer's frame header to the file/disk  */
        ierr = SLOG_STM_WriteFrameHdr( slog->frame_hdr );
        if ( ierr == SLOG_FAIL ) {
            fprintf( errfile, __FILE__":SLOG_AppendRecord() - "
                              "SLOG_STM_WriteFrameHdr( slog->Frame_Hdr ) "
                              "fails!!\n" );
            fflush( errfile );
            return SLOG_FAIL;
        };
        Nbytes_written += ierr;
          
if defined( DEBUG )
        slogfile_loc_cur = ftell( slog->fd );
        if ( slogfile_loc_cur != ( slog->frame_hdr ).inc_rec ) {
            fprintf( errfile, __FILE__":SLOG_AppendRecord() - "
                              "Inconsistency in File Pointers before calling "
                              "SLOG_STM_WriteBBuf() at line "__LINE__"\n" );
            fprintf( errfile, "       ""- slogfile_loc_cur = %d, but "
                              "( slog->frame_hdr ).inc_rec = %d\n",
                              slogfile_loc_cur, ( slog->frame_hdr ).inc_rec );
            fflush( errfile ); 
        };
#endif
      /*  Output all buffers' interval records to the file/disk  */
        ierr = SLOG_STM_WriteBBuf( &( slog->inc_bbuf ) );
        if ( ierr == SLOG_FAIL ) {
            fprintf( errfile, __FILE__":SLOG_AppendRecord() - "
                              "SLOG_STM_WriteBBuf( slog->INCOMMING_Bbuf ) "
                              "fails!!\n" );
            fflush( errfile );
            return SLOG_FAIL;
        };
        Nbytes_written += ierr;

if defined( DEBUG )
        slogfile_loc_cur = ftell( slog->fd );
        if ( slogfile_loc_cur != ( slog->frame_hdr ).cur_rec ) {
            fprintf( errfile, __FILE__":SLOG_AppendRecord() - "
                              "Inconsistency in File Pointers before calling "
                              "SLOG_STM_WriteBBuf() at line "__LINE__"\n" );
            fprintf( errfile, "       ""- slogfile_loc_cur = %d, but "
                              "( slog->frame_hdr ).cur_rec = %d\n",
                              slogfile_loc_cur, ( slog->frame_hdr ).cur_rec );
            fflush( errfile ); 
        };
#endif
        ierr = SLOG_STM_WriteBBuf( &( slog->cur_bbuf ) );
        if ( ierr == SLOG_FAIL ) {
            fprintf( errfile, __FILE__":SLOG_AppendRecord() - "
                              "SLOG_STM_WriteBBuf( slog->CURRENT_Bbuf ) "
                              "fails!!\n" );
            fflush( errfile );
            return SLOG_FAIL;
        };
        Nbytes_written += ierr;

      /*
          Pad the SLOGFILE in the unit of SLOG_BUFFER_UNIT byte block and
          Update the SLOG_STREAM's current slog file location 
      */
        fflush( slog->fd );
        ierr = fseek( slog->fd, slog->file_loc + SLOG_BUFFER_UNIT, SEEK_SET );
        slog->file_loc = ( SLOG_fptr ) ftell( slog->fd );

      /*
          Update the SLOG FILE's corresponding Frame Directory Entry
          by reading the SLOG_STREAM.file_dir_rec
      */  
        ierr = SLOG_WriteUpdatedFrameDirEntry( slog );
        ierr = SLOG_STM_IncrementFrameDirHdr( slog );
        ierr = SLOG_STM_WriteFrameDirHdr( slog );

      /*  Copy the Temporary Bbuf as the next Frame Buffer's Incomming Bbuf  */
        SLOG_Buf_CopyBBuf( &( slog->tmp_bbuf ), &( slog->inc_bbuf ) );
        SLOG_Buf_DelAllNodes( &( slog->cur_bbuf ) );
        SLOG_Buf_DelAllNodes( &( slog->out_bbuf ) );
        SLOG_Buf_DelAllNodes( &( slog->tmp_bbuf ) );

      /*  Reinitialize the Frame Header for next frame buffer  */
        SLOG_STM_UpdateFrameHdr( slog, 0, SLOG_fptr_NULL,
                                 SLOG_fptr_NULL, SLOG_fptr_NULL );

      /*  Reinitialize the bi-directional list buffers & related variables  */
        SLOG_Buf_Init( &( slog->cur_bbuf ) );
        SLOG_Buf_Init( &( slog->out_bbuf ) );
        SLOG_Buf_Init( &( slog->tmp_bbuf ) );
    };  /*  Endof { if ( new_buf_Nbytes_in_file > SLOG_BUFFER_UNIT ) }  */

#if defined( DEBUG )
  /*  Redundancy Check  */
  /*  Check if writing the input interval record to the buffer is possible  */
    new_buf_Nbytes_in_file = SLOG_typesz[ FrameHdr ]
                           + ( slog->inc_bbuf ).Nbytes_in_file
                           + ( slog->cur_bbuf ).Nbytes_in_file
                           + ( slog->out_bbuf ).Nbytes_in_file
                           + sizeof_IntvlRec_in_file;
    if ( new_buf_Nbytes_in_file > SLOG_BUFFER_UNIT ) {
        fprintf( errfile, __FILE__":SLOG_AppendRecord() - "
                          "The 2nd time in a single SLOG_AppendRecord() call "
                          "that new_buf_Nbytes_in_file (%d) > "
                          "SLOG_BUFFER_UNIT (%d) \n",
                          new_buf_Nbytes_in_file, SLOG_BUFFER_UNIT );
        fflush( errfile );
        return SLOG_FAIL;
    }
#endif
  /*
      since the new buffer size is NOT too big for the buffer 
      Add the new interval record to the buffer 
  */
    ierr = SLOG_Buf_AddIntvlRec( &( slog->cur_bbuf ), intvlrec );
    if ( ierr == SLOG_FAIL ) {
        fprintf( errfile, __FILE__":SLOG_AppendRecord() - "
                          "SLOG_Buf_AddIntvlRec() fails !!\n" );
        fflush( errfile );
        return SLOG_FAIL;
    };
}

int SLOG_CloseOutputStream( slog )
{
    int ierr;
}
